#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

"""

This file contains interfaces to the native pys file format.

PysReader and PysWriter classed are structured into the following sections:
 * shape
 * code
 * attributes
 * row_heights
 * col_widths
 * macros


**Provides**

 * :func:`wxcolor2rgb`
 * :dict:`wx2qt_fontweights`
 * :dict:`wx2qt_fontstyles`
 * :class:`PysReader`
 * :class:`PysWriter`

"""

from builtins import str, map, object

import ast
from base64 import b64decode, b85encode
from collections import OrderedDict
from typing import Any, BinaryIO, Callable, Iterable, Tuple

try:
    from pyspread.lib.attrdict import AttrDict
    from pyspread.lib.selection import Selection
    from pyspread.model.model import CellAttribute, CodeArray
except ImportError:
    from lib.attrdict import AttrDict
    from lib.selection import Selection
    from model.model import CellAttribute, CodeArray


def wxcolor2rgb(wxcolor: int) -> Tuple[int, int, int]:
    """Returns red, green, blue for given wxPython binary color value

    :param wxcolor: Color value from wx.Color

    """

    red = wxcolor >> 16
    green = wxcolor - (red << 16) >> 8
    blue = wxcolor - (red << 16) - (green << 8)

    return red, green, blue


wx2qt_fontweights = {
    90: 50,  # wx.FONTWEIGHT_NORMAL
    91: 25,  # wx.FONTWEIGHT_LIGHT
    92: 75,  # wx.FONTWEIGHT_BOLD
    93: 87,  # wx.FONTWEIGHT_MAX
    }

wx2qt_fontstyles = {
    90: 0,  # wx.FONTSTYLE_NORMAL
    93: 1,  # wx.FONTSTYLE_ITALIC
    94: 1,  # wx.FONTSTYLE_SLANT
    95: 2,  # wx.FONTSTYLE_MAX
    }


class PysReader:
    """Reads pys v2.0 file into a code_array"""

    def __init__(self, pys_file: BinaryIO, code_array: CodeArray):
        """
        :param pys_file: The pys or pysu file to be read
        :param code_array: Target code_array

        """

        self.pys_file = pys_file
        self.code_array = code_array

        self._section2reader = {
            "[Pyspread save file version]\n": self._pys_version,
            "[shape]\n": self._pys2shape,
            "[grid]\n": self._pys2code,
            "[attributes]\n": self._pys2attributes,
            "[row_heights]\n": self._pys2row_heights,
            "[col_widths]\n": self._pys2col_widths,
            "[macros]\n": self._pys2macros,
        }

        # When converting old versions, cell attributes are required that
        # take place after the cell attribute readout
        self.cell_attributes_postfixes = []

    def __iter__(self):
        """Iterates over self.pys_file, replacing everything in code_array"""

        state = None

        # Reset pys_file to start to enable multiple calls of this method
        self.pys_file.seek(0)

        for line in self.pys_file:
            line = line.decode("utf8")
            if line in self._section2reader:
                state = line
            elif state is not None:
                self._section2reader[state](line)
            yield line

        # Apply cell attributes post fixes
        for cell_attribute in self.cell_attributes_postfixes:
            self.code_array.cell_attributes.append(cell_attribute)

    # Decorators

    def version_handler(method: Callable) -> Callable:
        """Chooses method`_10` of method if version < 2.0

        :param method: Method to be replaced in case of old pys file version

        """

        def new_method(self, *args, **kwargs):
            if self.version <= 1.0:
                method10 = getattr(self, method.__name__+"_10")
                method10(*args, **kwargs)
            else:
                method(self, *args, **kwargs)

        return new_method

    # Helpers

    def _split_tidy(self, string: str, maxsplit: int = None) -> str:
        """Rstrips string for \n and splits string for \t

        :param string: String to be rstripped and  split
        :param maxsplit: Maximum number of splits

        """

        if maxsplit is None:
            return string.rstrip("\n").split("\t")
        else:
            return string.rstrip("\n").split("\t", maxsplit)

    def _get_key(self, *keystrings: str) -> Tuple[int, ...]:
        """Returns int key tuple from key string list

        :param keystrings: Strings that contain integers that are key elements

        """

        return tuple(map(int, keystrings))

    # Sections

    def _pys_version(self, line: str):
        """pys file version including assertion

        :param line: Pys file line to be parsed

        """

        self.version = float(line.strip())

        if self.version > 2.0:
            # Abort if file version not supported
            msg = "File version {version} unsupported (> 2.0)."
            raise ValueError(msg.format(version=line.strip()))

    def _pys2shape(self, line: str):
        """Updates shape in code_array

        :param line: Pys file line to be parsed

        """

        shape = self._get_key(*self._split_tidy(line))
        if any(dim <= 0 for dim in shape):
            # Abort if any axis is 0 or less
            msg = "Code array has invalid shape {shape}."
            raise ValueError(msg.format(shape=shape))
        self.code_array.shape = shape

    def _code_convert_1_2(self, key: Tuple[int, int, int], code: str) -> str:
        """Converts chart and image code from v1.0 to v2.0

        :param key: Key of cell with code
        :param code: Code in cell to be converted

        """

        def get_image_code(image_data: str, width: int, height: int) -> str:
            """Returns code string for v2.0

            :param image_data: b85encoded image data
            :param width: Image width
            :param height: Image height

            """

            image_buffer_tpl = 'bz2.decompress(base64.b85decode({data}))'
            image_array_tpl = 'numpy.frombuffer({buffer}, dtype="uint8")'
            image_matrix_tpl = '{array}.reshape({height}, {width}, 3)'

            image_buffer = image_buffer_tpl.format(data=image_data)
            image_array = image_array_tpl.format(buffer=image_buffer)
            image_matrix = image_matrix_tpl.format(array=image_array,
                                                   height=height, width=width)

            return image_matrix

        start_str = "bz2.decompress(base64.b64decode('"
        size_start_str = "wx.ImageFromData("
        if size_start_str in code and start_str in code:
            size_start = code.index(size_start_str) + len(size_start_str)
            size_str_list = code[size_start:].split(",")[:2]
            width, height = tuple(map(int, size_str_list))

            # We have a cell that displays a bitmap
            data_start = code.index(start_str) + len(start_str)
            data_stop = code.find("'", data_start)
            enc_data = bytes(code[data_start:data_stop], encoding='utf-8')
            compressed_image_data = b64decode(enc_data)
            reenc_data = b85encode(compressed_image_data)
            code = get_image_code(repr(reenc_data), width, height)

            selection = Selection([], [], [], [], [(key[0], key[1])])
            tab = key[2]
            attr_dict = AttrDict([("renderer", "image")])
            attr = CellAttribute(selection, tab, attr_dict)
            self.cell_attributes_postfixes.append(attr)

        elif "charts.ChartFigure(" in code:
            # We have a matplotlib figure
            selection = Selection([], [], [], [], [(key[0], key[1])])
            tab = key[2]
            attr_dict = AttrDict([("renderer", "matplotlib")])
            attr = CellAttribute(selection, tab, attr_dict)
            self.cell_attributes_postfixes.append(attr)

        return code

    def _pys2code_10(self, line: str):
        """Updates code in pys code_array - for save file version 1.0

        :param line: Pys file line to be parsed

        """

        row, col, tab, code = self._split_tidy(line, maxsplit=3)
        key = self._get_key(row, col, tab)

        if all(0 <= key[i] < self.code_array.shape[i] for i in range(3)):
            self.code_array.dict_grid[key] = str(self._code_convert_1_2(key,
                                                                        code))

    @version_handler
    def _pys2code(self, line: str):
        """Updates code in pys code_array

        :param line: Pys file line to be parsed

        """

        row, col, tab, code = self._split_tidy(line, maxsplit=3)
        key = self._get_key(row, col, tab)

        if all(0 <= key[i] < self.code_array.shape[i] for i in range(3)):
            self.code_array.dict_grid[key] = ast.literal_eval(code)

    def _attr_convert_1to2(self, key: str, value: Any) -> Tuple[str, Any]:
        """Converts key, value attribute pair from v1.0 to v2.0

        :param key: AttrDict key
        :param value: AttrDict value for key

        """

        color_attrs = ["bordercolor_bottom", "bordercolor_right", "bgcolor",
                       "textcolor"]
        if key in color_attrs:
            return key, wxcolor2rgb(value)

        elif key == "fontweight":
            return key, wx2qt_fontweights[value]

        elif key == "fontstyle":
            return key, wx2qt_fontstyles[value]

        elif key == "markup" and value:
            return "renderer", "markup"

        elif key == "angle" and value < 0:
            return "angle", 360 + value

        elif key == "merge_area":
            # Value in v1.0 None if the cell was merged
            # In v 2.0 this is no longer necessary
            return None, value

        # Update justifiaction and alignment values
        elif key in ["vertical_align", "justification"]:
            just_align_value_tansitions = {
                    "left": "justify_left",
                    "center": "justify_center",
                    "right": "justify_right",
                    "top": "align_top",
                    "middle": "align_center",
                    "bottom": "align_bottom",
            }
            return key, just_align_value_tansitions[value]

        return key, value

    def _pys2attributes_10(self, line: str):
        """Updates attributes in code_array - for save file version 1.0

        :param line: Pys file line to be parsed

        """

        splitline = self._split_tidy(line)

        selection_data = list(map(ast.literal_eval, splitline[:5]))
        selection = Selection(*selection_data)

        tab = int(splitline[5])

        attr_dict = AttrDict()

        old_merged_cells = {}

        for col, ele in enumerate(splitline[6:]):
            if not (col % 2):
                # Odd entries are keys
                key = ast.literal_eval(ele)

            else:
                # Even cols are values
                value = ast.literal_eval(ele)

                # Convert old wx color values and merged cells
                key_, value_ = self._attr_convert_1to2(key, value)

                if key_ is None and value_ is not None:
                    # We have a merged cell
                    old_merged_cells[value_[:2]] = value_
                try:
                    attr_dict.pop("merge_area")
                except KeyError:
                    pass
                attr_dict[key_] = value_

        attr = CellAttribute(selection, tab, attr_dict)
        self.code_array.cell_attributes.append(attr)

        for key in old_merged_cells:
            selection = Selection([], [], [], [], [key])
            attr_dict = AttrDict([("merge_area", old_merged_cells[key])])
            attr = CellAttribute(selection, tab, attr_dict)
            self.code_array.cell_attributes.append(attr)
        old_merged_cells.clear()

    @version_handler
    def _pys2attributes(self, line: str):
        """Updates attributes in code_array

        :param line: Pys file line to be parsed

        """

        splitline = self._split_tidy(line)

        selection_data = list(map(ast.literal_eval, splitline[:5]))
        selection = Selection(*selection_data)

        tab = int(splitline[5])

        attr_dict = AttrDict()

        for col, ele in enumerate(splitline[6:]):
            if not (col % 2):
                # Odd entries are keys
                key = ast.literal_eval(ele)

            else:
                # Even cols are values
                value = ast.literal_eval(ele)
                attr_dict[key] = value

        if attr_dict:  # Ignore empty attribute settings
            attr = CellAttribute(selection, tab, attr_dict)
            self.code_array.cell_attributes.append(attr)

    def _pys2row_heights(self, line: str):
        """Updates row_heights in code_array

        :param line: Pys file line to be parsed

        """

        # Split with maxsplit 3
        split_line = self._split_tidy(line)
        key = row, tab = self._get_key(*split_line[:2])
        height = float(split_line[2])

        shape = self.code_array.shape

        try:
            if row < shape[0] and tab < shape[2]:
                self.code_array.row_heights[key] = height

        except ValueError:
            pass

    def _pys2col_widths(self, line: str):
        """Updates col_widths in code_array

        :param line: Pys file line to be parsed

        """

        # Split with maxsplit 3
        split_line = self._split_tidy(line)
        key = col, tab = self._get_key(*split_line[:2])
        width = float(split_line[2])

        shape = self.code_array.shape

        try:
            if col < shape[1] and tab < shape[2]:
                self.code_array.col_widths[key] = width

        except ValueError:
            pass

    def _pys2macros(self, line: str):
        """Updates macros in code_array

        :param line: Pys file line to be parsed

        """

        self.code_array.macros += line


class PysWriter(object):
    """Interface between code_array and pys file data

    Iterating over it yields pys file lines

    """

    def __init__(self, code_array: CodeArray):
        """
        :param code_array: The code_array object data structure

        """

        self.code_array = code_array

        self.version = 2.0

        self._section2writer = OrderedDict([
            ("[Pyspread save file version]\n", self._version2pys),
            ("[shape]\n", self._shape2pys),
            ("[grid]\n", self._code2pys),
            ("[attributes]\n", self._attributes2pys),
            ("[row_heights]\n", self._row_heights2pys),
            ("[col_widths]\n", self._col_widths2pys),
            ("[macros]\n", self._macros2pys),
        ])

    def __iter__(self) -> Iterable[str]:
        """Yields a pys_file line wise from code_array"""

        for key in self._section2writer:
            yield key
            for line in self._section2writer[key]():
                yield line

    def __len__(self) -> int:
        """Returns how many lines will be written when saving the code_array"""

        lines = 9  # Headers + 1 line version + 1 line shape
        lines += len(self.code_array.dict_grid)
        lines += len(self.code_array.cell_attributes)
        lines += len(self.code_array.dict_grid.row_heights)
        lines += len(self.code_array.dict_grid.col_widths)
        lines += self.code_array.dict_grid.macros.count('\n')

        return lines

    def _version2pys(self) -> Iterable[str]:
        """Returns pys file version information in pys format

        Format: <version>\n

        """

        yield repr(self.version) + "\n"

    def _shape2pys(self) -> Iterable[str]:
        """Returns shape information in pys format

        Format: <rows>\t<cols>\t<tabs>\n

        """

        yield u"\t".join(map(str, self.code_array.shape)) + u"\n"

    def _code2pys(self) -> Iterable[str]:
        """Returns cell code information in pys format

        Format: <row>\t<col>\t<tab>\t<code>\n

        """

        for key in self.code_array:
            key_str = u"\t".join(repr(ele) for ele in key)
            if self.version <= 1.0:
                code_str = self.code_array(key)
            else:
                code_str = repr(self.code_array(key))
            out_str = key_str + u"\t" + code_str + u"\n"

            yield out_str

    def _attributes2pys(self) -> Iterable[str]:
        """Returns cell attributes information in pys format

        Format:
        <selection[0]>\t[...]\t<tab>\t<key>\t<value>\t[...]\n

        """

        # Remove doublettes
        purged_cell_attributes = []
        purged_cell_attributes_keys = []
        for selection, tab, attr_dict in self.code_array.cell_attributes:
            if purged_cell_attributes_keys and \
               (selection, tab) == purged_cell_attributes_keys[-1]:
                purged_cell_attributes[-1][2].update(attr_dict)
            else:
                purged_cell_attributes_keys.append((selection, tab))
                purged_cell_attributes.append([selection, tab, attr_dict])

        for selection, tab, attr_dict in purged_cell_attributes:
            if not attr_dict:
                continue

            sel_list = [selection.block_tl, selection.block_br,
                        selection.rows, selection.columns, selection.cells]

            tab_list = [tab]

            attr_dict_list = []
            for key in attr_dict:
                if key is not None:
                    attr_dict_list.append(key)
                    attr_dict_list.append(attr_dict[key])

            line_list = list(map(repr, sel_list + tab_list + attr_dict_list))

            yield u"\t".join(line_list) + u"\n"

    def _row_heights2pys(self) -> Iterable[str]:
        """Returns row height information in pys format

        Format: <row>\t<tab>\t<value>\n

        """

        for row, tab in self.code_array.dict_grid.row_heights:
            if row < self.code_array.shape[0] and \
               tab < self.code_array.shape[2]:
                height = self.code_array.dict_grid.row_heights[(row, tab)]
                height_strings = list(map(repr, [row, tab, height]))
                yield u"\t".join(height_strings) + u"\n"

    def _col_widths2pys(self) -> Iterable[str]:
        """Returns column width information in pys format

        Format: <col>\t<tab>\t<value>\n

        """

        for col, tab in self.code_array.dict_grid.col_widths:
            if col < self.code_array.shape[1] and \
               tab < self.code_array.shape[2]:
                width = self.code_array.dict_grid.col_widths[(col, tab)]
                width_strings = list(map(repr, [col, tab, width]))
                yield u"\t".join(width_strings) + u"\n"

    def _macros2pys(self) -> Iterable[str]:
        """Returns macros information in pys format

        Format: <macro code line>\n

        """

        macros = self.code_array.dict_grid.macros
        yield macros



from rpy2 import robjects
from rpy2.interactive import process_revents
from rpy2.robjects.packages import importr, data
from rpy2.robjects.vectors import IntVector, FloatVector
from rpy2.robjects.lib import grdevices

try:
    process_revents.start()
except:
    pass

base = importr('base')
lattice = importr('lattice')
datasets = importr('datasets')

rprint = robjects.globalenv.find("print")

volcano = data(datasets).fetch("volcano")["volcano"]
plot = lattice.wireframe(volcano, shade = True,
                         zlab = "value",
                         aspect = FloatVector((61.0/87, 0.4)),
                         light_source = IntVector((10,0,10)))

with grdevices.render_to_bytesio(grdevices.svg) as svg:
    rprint(plot)
process_revents.stop()

svg.getvalue() 



fig = Figure()
ax = fig.add_axes([.2,.2, .7, .7])

numpy.random.seed(19680801)
number_of_bins = 20

# An example of three data sets to compare
number_of_data_points = 387
labels = ["A", "B", "C"]
data_sets = [numpy.random.normal(0, 1, number_of_data_points),
             numpy.random.normal(6, 1, number_of_data_points),
             numpy.random.normal(-3, 1, number_of_data_points)]

# Computed quantities to aid plotting
hist_range = (numpy.min(data_sets), numpy.max(data_sets))
binned_data_sets = []
for d in data_sets:
    h = numpy.histogram(d, range=hist_range, bins=number_of_bins)[0]
    binned_data_sets.append(h)

binned_maximums = numpy.max(binned_data_sets, axis=1)
x_locations = numpy.arange(0, sum(binned_maximums), numpy.max(binned_maximums))

# The bin_edges are the same for all of the histograms
bin_edges = numpy.linspace(hist_range[0], hist_range[1], number_of_bins + 1)
centers = 0.5 * (bin_edges + numpy.roll(bin_edges, 1))[:-1]
heights = numpy.diff(bin_edges)

# Cycle through and plot each histogram
for x_loc, binned_data in zip(x_locations, binned_data_sets):
    lefts = x_loc - 0.5 * binned_data
    ax.barh(centers, binned_data, height=heights, left=lefts)

ax.set_xticks(x_locations)
ax.set_xticklabels(labels)

ax.set_ylabel("Data values")
ax.set_xlabel("Data sets")

ax.set_title('Multiple vertical histogram charts', pad=15)

fig



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

"""

**Provides**

 * :func:`sniff`: Sniffs CSV dialect and header info
 * :func:`get_header`
 * :func:`csv_reader`
 * :func:`convert`
 * :func:`date`
 * :func:`datetime`
 * :func:`time`
 * :func:`make_object`
 * :dict:`typehandlers`

"""

import ast
import csv
from decimal import Decimal
from pathlib import Path
from typing import TextIO, Iterable, List

try:
    from dateutil.parser import parse
except ImportError:
    parse = None

try:
    from moneyed import Money, list_all_currencies
except ImportError:
    Money = None


def sniff(filepath: Path, sniff_size: int, encoding: str) -> csv.Dialect:
    """Sniffs CSV dialect and header info

    :param filepath: Path of file to sniff
    :param sniff_size: Maximum no. bytes to use for sniffing
    :param encoding: File encoding
    :return: csv.Dialect object with additional attribute `has_header`

    """

    with open(filepath, newline='', encoding=encoding) as csvfile:
        csv_str = csvfile.read(sniff_size)

    dialect = csv.Sniffer().sniff(csv_str)
    setattr(dialect, "hasheader", csv.Sniffer().has_header(csv_str))
    setattr(dialect, "encoding", encoding)
    if dialect.escapechar is None:
        setattr(dialect, "quoting", csv.QUOTE_NONE)
        setattr(dialect, "escapechar", '"')

    return dialect


def get_header(csvfile: TextIO, dialect: csv.Dialect) -> List[str]:
    """Returns list of first line items of file filepath

    :param csvfile: CSV file
    :param dialect: Dialect of CSV file

    """

    csvfile.seek(0)
    csvreader = csv.reader(csvfile, dialect=dialect)
    for header in csvreader:
        break

    csvfile.seek(0)
    return header


def csv_reader(csvfile: TextIO, dialect: csv.Dialect) -> Iterable[str]:
    """Generator of str values from csv file in filepath, ignores header

    :param csvfile: Csv file to read
    :param dialect: Csv dialect

    """

    csvreader = csv.reader(csvfile, dialect=dialect)
    try:
        ignore_header = dialect.hasheader and not dialect.keepheader
    except AttributeError:
        try:
            ignore_header = dialect.hasheader
        except AttributeError:
            ignore_header = False

    if ignore_header:
        for line in csvreader:
            break

    for line in csvreader:
        yield line


# Type conversion functions

def convert(string: str, digest_type: str) -> str:
    """Main type conversion function for csv import

    :param string: String to be digested
    :param digest_type: Name of digetsion function
    :return: Converted string

    """

    if digest_type is None:
        digest_type = 'repr'

    if digest_type.split()[0] == "Money":
        currency = digest_type.split()[1][1:-1]
        return repr(typehandlers["Money"](string, currency=currency))

    try:
        return repr(typehandlers[digest_type](string))

    except Exception:
        return repr(string)


def date(obj):
    """Makes a date from comparable types"""

    try:
        return parse(obj).date()
    except TypeError as err:
        return err


def datetime(obj):
    """Makes a datetime from comparable types"""

    try:
        return parse(obj)
    except TypeError as err:
        return err


def time(obj):
    """Makes a time from comparable types"""

    try:
        return parse(obj).time()
    except TypeError as err:
        return err


def make_object(obj):
    """Parses the object with ast.literal_eval"""

    return ast.literal_eval(obj)


typehandlers = {
    'object': ast.literal_eval,
    'repr': lambda x: x,
    'str': str,
    'int': int,
    'float': float,
    'decimal': Decimal,
    'complex': complex,
    'bool': bool,
    'bytes': bytes,
}


if Money is not None:
    def money(amount, currency="USD"):
        """Money wrapper defining a standard currency"""

        return Money(amount, currency=currency)

    typehandlers.update({'Money': money})

    currencies = list_all_currencies()
else:
    currencies = []


if parse is not None:
    typehandlers.update({'date': date,
                         'datetime': datetime,
                         'time': time,
                         })



from matplotlib import cm
from mpl_toolkits.mplot3d import Axes3D

fig = Figure()
ax = fig.add_subplot(111, projection='3d')

X = numpy.arange(-5, 5, 0.25)
Y = numpy.arange(-5, 5, 0.25)
X, Y = numpy.meshgrid(X, Y)
R = numpy.sqrt(X**2 + Y**2)
Z = numpy.sin(R)

ax.plot_surface(X, Y, Z, rstride=1, cstride=1, cmap=cm.viridis)

ax.set_xlabel('X', labelpad=5)
ax.set_ylabel('Y', labelpad=5)
ax.set_zlabel('Z', labelpad=5)
ax.tick_params('x', pad=0)
ax.tick_params('y', pad=0)
ax.tick_params('z', pad=3)

ax.set_title("Surface chart", pad=20)

fig


# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------


"""
test_hashing
============

Unit tests for hashing.py

"""

import pytest
from ..hashing import genkey, sign, verify


KEYS = [genkey() for _ in range(100)]


def test_genkey():
    """Unit test for genkey"""

    keys = []

    for length in range(1, 128):
        keys.append(genkey(length))

    for i, key in enumerate(keys):
        assert len(key) == i + 1

    assert len(set(KEYS)) == len(KEYS)


param_test_sign = [
    ("", None, "", ValueError),
    ("2", "123", b"", TypeError),
    (b"2", b"123", b"287715fdf85b987e2bc3f468b4a53346aee8da24305ff4456dd3390c"
                   b"7ab5a7cdf5f5dbb6831310880909aedfd21e59f1b7c4421482a33b3"
                   b"450c38ffcbf1bf11e", None),
    ("2", "123", b"", TypeError),
    (b"2", 123, "", ValueError),
    (b"2", b"1"*1000, b"", UserWarning),
]

@pytest.mark.parametrize("data, key, res, error", param_test_sign)
def test_sign(data, key, res, error):
    """Unit test for sign"""
    if error and issubclass(error, Exception):
        with pytest.raises(error):
            sign(data, key)
    elif error and issubclass(error, UserWarning):
        with pytest.warns(error):
            sign(data, key)
    else:
        assert sign(data, key) ==  res

param_test_sign_verify = [
    (b"Test", KEYS[0], b"Test", KEYS[0], True),
    (100*"\u2200".encode('utf-8'), KEYS[0], 100*"\u2200".encode('utf-8'),
     KEYS[0], True),
    (b"Test", KEYS[0], b"Test", KEYS[1], False),
    (b"Test", KEYS[0], b"TEST", KEYS[0], False),
    (b"Test", KEYS[0], b"TEST", KEYS[3], False),
    (b"", KEYS[0], b"", KEYS[3], False),
    (b"", KEYS[0], b"", KEYS[0], True),
    (b"Hello World\n"*10, KEYS[1], b"Hello World\n"*10, KEYS[1], True),
]


@pytest.mark.parametrize("data1, sigkey, data2, verkey, res",
                         param_test_sign_verify)
def test_sign_verify(data1, sigkey, data2, verkey, res):
    """Unit test for sign and verify"""

    signature = sign(data1, sigkey)
    assert verify(data2, signature, verkey) == res



#!/usr/bin/python3
# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

try:
    from .pyspread import main
except ImportError:
    from pyspread import main

main()



# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------


"""
test_workflows
==============

Unit tests for workflows.py

"""

from contextlib import contextmanager
from os.path import abspath, dirname, join
from pathlib import Path
import sys

import pytest

from PyQt5.QtCore import Qt, QItemSelectionModel
from PyQt5.QtWidgets import QApplication

try:
    from pyspread.dialogs import GridShapeDialog
except ImportError:
    from dialogs import GridShapeDialog


PYSPREADPATH = abspath(join(dirname(__file__) + "/.."))
LIBPATH = abspath(PYSPREADPATH + "/lib")


@contextmanager
def insert_path(path):
    sys.path.insert(0, path)
    yield
    sys.path.pop(0)


with insert_path(PYSPREADPATH):
    from ..pyspread import MainWindow


app = QApplication.instance()
if app is None:
    app = QApplication([])
main_window = MainWindow()


class TestWorkflows:
    """Unit tests for Workflows in workflows.py"""

    workflows = main_window.workflows

    def test_busy_cursor(self):
        """Unit test for busy_cursor"""

        assert QApplication.overrideCursor() != Qt.WaitCursor

        with self.workflows.busy_cursor():
            assert QApplication.overrideCursor() == Qt.WaitCursor

        assert QApplication.overrideCursor() != Qt.WaitCursor

    def test_prevent_updates(self):
        """Unit test for prevent_updates"""

        assert not main_window.prevent_updates

        with self.workflows.prevent_updates():
            assert main_window.prevent_updates

        assert not main_window.prevent_updates

    def test_reset_changed_since_save(self):
        """Unit test for reset_changed_since_save"""

        main_window.settings.changed_since_save = True
        self.workflows.reset_changed_since_save()
        assert not main_window.settings.changed_since_save

    param_update_main_window_title = [
        (Path.home(), "pyspread"),
        (Path("/test.pys"), "test.pys - pyspread"),
    ]

    @pytest.mark.parametrize("path, title", param_update_main_window_title)
    def test_update_main_window_title(self, path, title):
        """Unit test for update_main_window_title"""

        main_window.settings.last_file_input_path = path
        self.workflows.update_main_window_title()
        assert main_window.windowTitle() == title

    param_file_new = [
        ((1000, 100, 3), (1000, 100, 3), None),
        ((100, 100, 3), (100, 100, 3), None),
        ((10000000, 100, 3), (1000, 100, 3),
         "Error: Grid shape (10000000, 100, 3) exceeds (1000000, 100000, 100)."
         ),
        (None, (1000, 100, 3), None),
    ]

    @pytest.mark.parametrize("shape, res, msg", param_file_new)
    def test_file_new(self, shape, res, msg, monkeypatch):
        """Unit test for file_new"""

        monkeypatch.setattr(GridShapeDialog, "shape", shape)
        self.workflows.file_new()

        assert main_window.grid.model.shape == res
        assert main_window.grid.current == (0, 0, 0)
        assert main_window.settings.last_file_input_path == Path.home()
        assert main_window.settings.changed_since_save is False
        assert main_window.safe_mode is False
        if msg:
            assert main_window.statusBar().currentMessage() == msg

        monkeypatch.setattr(GridShapeDialog, "shape",
                            main_window.settings.shape)
        self.workflows.file_new()

    param_count_file_lines = [
        ("", 0, "counttest.txt", None),
        ("\n"*100, 100, "counttest.txt", None),
        ("Test"*100, 0, "counttest.txt", None),
        ("Test\n"*10, 10, "counttest.txt", None),
        ("Test\n"*10, None, "false_filename.txt", "Error"),
    ]

    @pytest.mark.parametrize("txt, res, filename, msg", param_count_file_lines)
    def test_count_file_lines(self, txt, res, filename, msg, tmpdir):
        """Unit test for count_file_lines"""

        tmpfile = tmpdir / "counttest.txt"
        tmpfile.write_text(txt, "utf-8")
        testfile = tmpdir / filename
        assert self.workflows.count_file_lines(testfile) == res
        if msg:
            assert str(testfile) in main_window.statusBar().currentMessage()
        tmpfile.remove()

    def test_edit_sort_ascending(self):
        """Unit test for test_edit_sort_ascending"""

        main_window.grid.model.code_array[0, 0, 0] = "1"
        main_window.grid.model.code_array[1, 0, 0] = "3"
        main_window.grid.model.code_array[2, 0, 0] = "2"
        main_window.grid.model.code_array[0, 1, 0] = "12"
        main_window.grid.model.code_array[1, 1, 0] = "33"
        main_window.grid.model.code_array[2, 1, 0] = "24"

        for row in range(3):
            for column in range(2):
                index = main_window.grid.model.index(row, column)
                main_window.grid.selectionModel().select(
                    index, QItemSelectionModel.Select)

        self.workflows.edit_sort_ascending()
        assert main_window.grid.model.code_array((0, 0, 0)) == "1"
        assert main_window.grid.model.code_array((1, 0, 0)) == "2"
        assert main_window.grid.model.code_array((2, 1, 0)) == "33"

    def test_edit_sort_descending(self):
        """Unit test for test_edit_sort_descending"""

        main_window.grid.model.code_array[0, 0, 0] = "1"
        main_window.grid.model.code_array[1, 0, 0] = "3"
        main_window.grid.model.code_array[2, 0, 0] = "2"
        main_window.grid.model.code_array[0, 1, 0] = "12"
        main_window.grid.model.code_array[1, 1, 0] = "33"
        main_window.grid.model.code_array[2, 1, 0] = "24"

        for row in range(3):
            for column in range(2):
                index = main_window.grid.model.index(row, column)
                main_window.grid.selectionModel().select(
                    index, QItemSelectionModel.Select)

        self.workflows.edit_sort_descending()
        assert main_window.grid.model.code_array((0, 0, 0)) == "3"
        assert main_window.grid.model.code_array((1, 0, 0)) == "2"
        assert main_window.grid.model.code_array((2, 1, 0)) == "12"



fig = Figure()
ax = fig.add_axes([.2,.05, .7, .7])

category_names = ['Strongly disagree', 'Disagree',
                  'Neither agree nor disagree', 'Agree', 'Strongly agree']
results = {
    'Question 1': [10, 15, 17, 32, 26],
    'Question 2': [26, 22, 29, 10, 13],
    'Question 3': [35, 37, 7, 2, 19],
    'Question 4': [32, 11, 9, 15, 33],
    'Question 5': [21, 29, 5, 5, 40],
    'Question 6': [8, 19, 5, 30, 38]
}


def survey(fig, ax, results, category_names):
    """
    Parameters
    ----------
    results : dict
        A mapping from question labels to a list of answers per category.
        It is assumed all lists contain the same number of entries and that
        it matches the length of *category_names*.
    category_names : list of str
        The category labels.
    """

    from matplotlib import cm
    labels = list(results.keys())
    data = numpy.array(list(results.values()))
    data_cum = data.cumsum(axis=1)
    category_colors = cm.RdYlGn(numpy.linspace(0.15, 0.85, data.shape[1]))

    ax.invert_yaxis()
    ax.xaxis.set_visible(False)
    ax.set_xlim(0, numpy.sum(data, axis=1).max())

    for i, (colname, color) in enumerate(zip(category_names, category_colors)):
        widths = data[:, i]
        starts = data_cum[:, i] - widths
        ax.barh(labels, widths, left=starts, height=0.5,
                label=colname, color=color)
        xcenters = starts + widths / 2

        r, g, b, _ = color
        text_color = 'white' if r * g * b < 0.5 else 'darkgrey'
        for y, (x, c) in enumerate(zip(xcenters, widths)):
            ax.text(x, y, str(int(c)), ha='center', va='center',
                    color=text_color)
    ax.legend(ncol=2, bbox_to_anchor=(0, 1),
              loc='lower left', fontsize='small')

    ax.set_title("Normalized stacked bar chart", pad=60)

    return fig


survey(fig, ax, results, category_names)



fig = Figure()
ax = fig.add_axes([.2,.2, .7, .7])

# Data for plotting
t = numpy.arange(0.0, 2.0, 0.01)
s = 1 + numpy.sin(2 * numpy.pi * t)

ax.plot(t, s)
ax.fill_between(t, 0, s, facecolor='blue', alpha=0.5)

ax.set_xlim(0, 2)
ax.set_ylim(0, None)

ax.set(xlabel='Time (s)', ylabel='Voltage (mV)',
       title='Area chart')
ax.grid()

fig



fig = Figure()
ax = fig.add_axes([.2,.2, .7, .7])

N = 5
menMeans = (20, 35, 30, 35, 27)
womenMeans = (25, 32, 34, 20, 25)
menStd = (2, 3, 4, 1, 2)
womenStd = (3, 5, 2, 3, 3)
ind = numpy.arange(N)    # the x locations for the groups
width = 0.35       # the width of the bars: can also be len(x) sequence


p1 = ax.bar(ind, menMeans, width, yerr=menStd)
p2 = ax.bar(ind, womenMeans, width,
             bottom=menMeans, yerr=womenStd)

ax.yaxis.set_label('Scores')
ax.set_title('Stacked column chart')
ax.set_xticks(ind)
ax.set_xticklabels(('G1', 'G2', 'G3', 'G4', 'G5'))
ax.set_yticks(numpy.arange(0, 81, 10))
ax.legend((p1[0], p2[0]), ('Men', 'Women'))

fig


# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

"""
Pyspread undoable commands

**Provides**

* :class:`SetGridSize`
* :class:`SetCellCode`
* :class:`SetCellFormat`
* :class:`SetCellMerge`
* :class:`SetCellRenderer`
* :class:`SetCellTextAlignment`
* :class:`SetColumnWidth`
* :class:`SetRowHeight`


"""

from copy import copy
from itertools import cycle
from math import isclose
from typing import List, Iterable, Tuple

from PyQt5.QtCore import Qt, QModelIndex, QAbstractTableModel
from PyQt5.QtGui import QTextDocument
from PyQt5.QtWidgets import QUndoCommand, QTableView, QPlainTextEdit

try:
    from pyspread.model.model import CellAttribute

    from pyspread.lib.attrdict import AttrDict
    from pyspread.lib.selection import Selection
except ImportError:
    from model.model import CellAttribute

    from lib.attrdict import AttrDict
    from lib.selection import Selection


class SetGridSize(QUndoCommand):
    """Sets size of grid"""

    def __init__(self, grid: QTableView, old_shape: Tuple[int, int, int],
                 new_shape: Tuple[int, int, int], description: str):
        """
        :param grid: The main grid object
        :param old_shape: Shape of the grid before command
        :param new_shape: Shape of the grid to be set
        :param description: Command description

        """

        super().__init__(description)

        self.grid = grid
        self.old_shape = old_shape
        self.new_shape = new_shape

        self.deleted_cells = {}  # Storage dict for deleted cells

    def redo(self):
        """Redo grid size change and deletion of cell code outside new shape

        Cell formats are not deleted.

        """

        model = self.grid.model
        code_array = model.code_array

        rows, columns, tables = self.new_shape
        shapeselection = Selection([(0, 0)], [(rows-1, columns-1)], [], [], [])

        for row, column, table in code_array.keys():
            if not (table < tables and (row, column) in shapeselection):
                # Code outside grid shape. Delete it and store cell data
                key = row, column, table
                self.deleted_cells[key] = code_array.pop(key)

        # Now change the shape
        self.grid.model.shape = self.new_shape

    def undo(self):
        """Undo grid size change and deletion of cell code outside new shape

        Cell formats are not deleted.

        """
        model = self.grid.model

        model.shape = self.old_shape

        for row, column, table in self.deleted_cells:
            index = model.index(row, column, QModelIndex())
            code = self.deleted_cells[(row, column, table)]
            model.setData(index, code, Qt.EditRole, raw=True, table=table)


class PasteSelectedCellData(QUndoCommand):
    """Paste selected cells"""

    def __init__(self, grid: QTableView, model: QAbstractTableModel,
                 selection: Selection, data: str, description: str):
        """
        :param grid: The main grid object
        :param model: Model of the grid object
        :param selection: Selection into which cells are pasted
        :param data: Data to be pasted
        :param description: Command description

        """
        super().__init__(description)
        self.grid = grid
        self.model = model
        self.selection = selection
        self.data = data

    def redo(self):
        """Redo cell data deletion, updates screen"""

        self.old_code = {}

        code_array = self.model.code_array
        shape = self.model.shape
        table = self.grid.table
        selection = self.selection

        (top, left), (bottom, right) = selection.get_grid_bbox(shape)

        paste_gen = (line.split("\t") for line in self.data.split("\n"))
        for row, line in enumerate(cycle(paste_gen)):
            paste_row = row + top
            if paste_row > bottom or (paste_row, 0, table) not in code_array:
                break
            for column, value in enumerate(cycle(line)):
                paste_column = column + left
                paste_key = paste_row, paste_column, table
                if (paste_key in code_array
                        and paste_column <= right):
                    if ((paste_row, paste_column) in selection and not
                            code_array.cell_attributes[paste_key].locked):
                        key = paste_row, paste_column, table
                        # Preserve line breaks
                        value = value.replace("\u000C", "\n")
                        try:
                            self.old_code[key] = code_array(key)
                            code_array[key] = value
                        except IndexError:
                            pass
                else:
                    break

        self.model.dataChanged.emit(QModelIndex(), QModelIndex())

    def undo(self):
        """Undo row insertion, updates screen"""

        for key in self.old_code:
            self.model.code_array[key] = self.old_code[key]
        self.old_code.clear()
        self.model.dataChanged.emit(QModelIndex(), QModelIndex())


class SetCellCode(QUndoCommand):
    """Sets cell code in grid"""

    def __init__(self, code: str, model: QAbstractTableModel,
                 index: QModelIndex, description: str):
        """
        :param code: The main grid object
        :param model: Model of the grid object
        :param index: Index of the cell for which the code is set
        :param description: Command description

        """
        super().__init__(description)

        self.description = description
        self.model = model
        self.indices = [index]
        self.old_codes = [model.code(index)]
        self.new_codes = [code]

    def id(self):
        return 1  # Enable command merging

    def mergeWith(self, other: QUndoCommand) -> bool:
        """Consecutive commands are merged if descriptions match

        :param other: Command to be merged

        """

        if self.description != other.description:
            return False
        self.new_codes += other.new_codes
        self.old_codes += other.old_codes
        self.indices += other.indices
        return True

    def redo(self):
        """Redo cell code setting

        During update, cell highlighting is disabled.

        """

        with self.model.main_window.entry_line.disable_updates():
            for index, new_code in zip(self.indices, self.new_codes):
                self.model.setData(index, new_code, Qt.EditRole, raw=True)
        self.model.dataChanged.emit(QModelIndex(), QModelIndex())

    def undo(self):
        """Undo cell code setting.

        During update, cell highlighting is disabled.

        """

        with self.model.main_window.entry_line.disable_updates():
            for index, old_code in zip(self.indices, self.old_codes):
                self.model.setData(index, old_code, Qt.EditRole, raw=True)
        self.model.dataChanged.emit(QModelIndex(), QModelIndex())


class SetRowsHeight(QUndoCommand):
    """Sets rows height in grid"""

    def __init__(self, grid: QTableView, rows: List[int], table: int,
                 old_height: float, new_height: float, description: str):
        """
        :param grid: The main grid object
        :param rows: Rows for which height are set
        :param table: Table for which row heights are set
        :param old_height: Row height before setting
        :param new_height: Target row height for setting
        :param description: Command description

        """

        super().__init__(description)

        self.grid = grid
        self.rows = rows
        self.table = table
        self.old_height = old_height
        self.new_height = new_height

        self.default_size = \
            self.grid.verticalHeader().defaultSectionSize() / self.grid.zoom

    def id(self) -> int:
        """Command id that enables command merging"""

        return 2

    def mergeWith(self, other: QUndoCommand) -> bool:
        """Consecutive commands are merged if descriptions match

        :param other: Command to be merged

        """

        if self.rows != other.rows:
            return False
        self.new_height = other.new_height
        return True

    def redo(self):
        """Redo row height setting"""

        # Update all frontend grids
        for grid in self.grid.main_window.grids:
            with grid.undo_resizing_row():
                for row in self.rows:
                    grid.setRowHeight(row, int(self.new_height * grid.zoom))

        code_array = self.grid.model.code_array
        with self.grid.undo_resizing_row():
            for row in self.rows:
                if isclose(self.new_height, self.default_size):
                    try:
                        code_array.row_heights.pop((row, self.table))
                    except KeyError:
                        pass
                else:
                    code_array.row_heights[(row, self.table)] = self.new_height

    def undo(self):
        """Undo row height setting"""

        # Update all frontend grids
        for grid in self.grid.main_window.grids:
            with grid.undo_resizing_row():
                for row in self.rows:
                    grid.setRowHeight(row, int(self.old_height * grid.zoom))

        code_array = self.grid.model.code_array
        with self.grid.undo_resizing_row():
            for row in self.rows:
                if isclose(self.old_height, self.default_size):
                    try:
                        code_array.row_heights.pop((row, self.table))
                    except KeyError:
                        pass
                else:
                    code_array.row_heights[(row, self.table)] = self.old_height


class SetColumnsWidth(QUndoCommand):
    """Sets column width in grid"""

    def __init__(self, grid: QTableView, columns: List[int], table: int,
                 old_width: float, new_width: float, description: str):
        """
        :param grid: The main grid object
        :param columns: Columns for which widths are set
        :param table: Table for which column widths are set
        :param old_width: Column width before setting
        :param new_width: Target column width for setting
        :param description: Command description

        """

        super().__init__(description)

        self.grid = grid
        self.columns = columns
        self.table = table
        self.old_width = old_width
        self.new_width = new_width

        self.default_size = \
            self.grid.horizontalHeader().defaultSectionSize() / self.grid.zoom

    def id(self) -> int:
        """Command id that enables command merging"""

        return 3  # Enable command merging

    def mergeWith(self, other: QUndoCommand) -> bool:
        """Consecutive commands are merged if descriptions match

        :param other: Command to be merged

        """

        if self.columns != other.columns:
            return False
        self.new_width = other.new_width
        return True

    def redo(self):
        """Redo column width setting"""

        # Update all frontend grids
        for grid in self.grid.main_window.grids:
            with grid.undo_resizing_column():
                for column in self.columns:
                    grid.setColumnWidth(column,
                                        int(self.new_width * grid.zoom))

        code_array = self.grid.model.code_array
        with self.grid.undo_resizing_column():
            for column in self.columns:
                if isclose(self.new_width, self.default_size):
                    try:
                        code_array.col_widths.pop((column, self.table))
                    except KeyError:
                        pass
                else:
                    code_array.col_widths[(column, self.table)] = \
                        self.new_width

    def undo(self):
        """Undo column width setting"""

        # Update all frontend grids
        for grid in self.grid.main_window.grids:
            with grid.undo_resizing_column():
                for column in self.columns:
                    grid.setColumnWidth(column, int(self.old_width*grid.zoom))

        code_array = self.grid.model.code_array
        with self.grid.undo_resizing_column():
            for column in self.columns:
                if isclose(self.old_width, self.default_size):
                    try:
                        code_array.col_widths.pop((column, self.table))
                    except KeyError:
                        pass
                else:
                    code_array.col_widths[(column, self.table)] = \
                        self.old_width


class DeleteSelectedCellData(QUndoCommand):
    """Delete selected cells"""

    def __init__(self, grid: QTableView, model: QAbstractTableModel,
                 selection: Selection, description: str):
        """
        :param grid: The main grid object
        :param model: Model of the grid object
        :param selection: Selected cells to be deleted
        :param description: Command description

        """
        super().__init__(description)
        self.grid = grid
        self.model = model
        self.selection = selection

    def redo(self):
        """Redo cell data deletion, updates screen"""

        self.old_code = {}
        for key in self.selection.cell_generator(self.model.shape,
                                                 self.grid.table):
            if not self.model.code_array.cell_attributes[key]['locked']:
                try:
                    self.old_code[key] = self.model.code_array.pop(key)
                except KeyError:
                    pass

        self.model.dataChanged.emit(QModelIndex(), QModelIndex())

    def undo(self):
        """Undo row insertion, updates screen"""

        for key in self.old_code:
            self.model.code_array[key] = self.old_code[key]
        self.old_code.clear()
        self.model.dataChanged.emit(QModelIndex(), QModelIndex())


class InsertRows(QUndoCommand):
    """Inserts grid rows"""

    def __init__(self, grid: QTableView, model: QAbstractTableModel,
                 index: QModelIndex, row: int, count: int, description: str):
        """
        :param grid: The main grid object
        :param model: Model of the grid object
        :param index: Parent into which the new rows are inserted
        :param row: Row number that first row will have after insertion
        :param count: Number of rows to be inserted
        :param description: Command description

        """

        super().__init__(description)
        self.grid = grid
        self.model = model
        self.index = index
        self.first = self.row = row
        self.last = row + count
        self.count = count

    def redo(self):
        """Redo row insertion, updates screen"""

        # Store content of overflowing rows
        self.old_row_heights = copy(self.model.code_array.row_heights)
        self.old_cell_attributes = copy(self.model.code_array.cell_attributes)
        self.old_code = {}

        no_rows = self.model.shape[0]
        rows = list(range(no_rows-self.count, no_rows+1))
        selection = Selection([], [], rows, [], [])
        for key in selection.cell_generator(self.model.shape, self.grid.table):
            old_code = self.model.code_array(key)
            if old_code is not None:
                self.old_code[key] = old_code

        with self.model.inserting_rows(self.index, self.first, self.last):
            self.model.insertRows(self.row, self.count)
        self.grid.table_choice.on_table_changed(self.grid.table)

    def undo(self):
        """Undo row insertion, updates screen"""

        # Clear must be first so that merged cells do not consume values
        self.model.code_array.dict_grid.cell_attributes.clear()

        with self.model.removing_rows(self.index, self.first, self.last):
            self.model.removeRows(self.row, self.count)

        self.model.code_array.dict_grid.row_heights = self.old_row_heights

        for ca in self.old_cell_attributes:
            self.model.code_array.dict_grid.cell_attributes.append(ca)

        for key in self.old_code:
            self.model.code_array[key] = self.old_code[key]
        self.grid.table_choice.on_table_changed(self.grid.table)


class DeleteRows(QUndoCommand):
    """Deletes grid rows"""

    def __init__(self, grid: QTableView, model: QAbstractTableModel,
                 index: QModelIndex, row: int, count: int, description: str):
        """
        :param grid: The main grid object
        :param model: Model of the grid object
        :param index: Parent from which the new rows are deleted
        :param row: Row number of the first row to be deleted
        :param count: Number of rows to be deleted
        :param description: Command description

        """

        super().__init__(description)
        self.grid = grid
        self.model = model
        self.index = index
        self.first = self.row = row
        self.last = row + count
        self.count = count

    def redo(self):
        """Redo row deletion, updates screen"""

        # Store content of deleted rows
        self.old_row_heights = copy(self.model.code_array.row_heights)
        self.old_cell_attributes = copy(self.model.code_array.cell_attributes)
        self.old_code = {}
        rows = list(range(self.first, self.last+1))
        selection = Selection([], [], rows, [], [])
        for key in selection.cell_generator(self.model.shape, self.grid.table):
            self.old_code[key] = self.model.code_array(key)

        with self.model.removing_rows(self.index, self.first, self.last):
            self.model.removeRows(self.row, self.count)
        self.grid.table_choice.on_table_changed(self.grid.table)

    def undo(self):
        """Undo row deletion, updates screen"""

        # Clear must be first so that merged cells do not consume values
        self.model.code_array.dict_grid.cell_attributes.clear()

        with self.model.inserting_rows(self.index, self.first, self.last):
            self.model.insertRows(self.row, self.count)

        self.model.code_array.dict_grid.row_heights = self.old_row_heights

        for ca in self.old_cell_attributes:
            self.model.code_array.dict_grid.cell_attributes.append(ca)
        for key in self.old_code:
            self.model.code_array[key] = self.old_code[key]

        self.grid.table_choice.on_table_changed(self.grid.table)


class InsertColumns(QUndoCommand):
    """Inserts grid columns"""

    def __init__(self, grid: QTableView, model: QAbstractTableModel,
                 index: QModelIndex, column: int, count: int,
                 description: str):
        """
        :param grid: The main grid object
        :param model: Model of the grid object
        :param index: Parent into which the new columns are inserted
        :param column: Column number of the first column after insertion
        :param count: Number of columns to be inserted
        :param description: Command description

        """

        super().__init__(description)
        self.grid = grid
        self.model = model
        self.index = index
        self.column = column
        self.first = self.column = column
        self.last = column + count
        self.count = count

    def redo(self):
        """Redo column insertion, updates screen"""

        # Store content of overflowing columns
        self.old_col_widths = copy(self.model.code_array.col_widths)
        self.old_cell_attributes = copy(self.model.code_array.cell_attributes)
        self.old_code = {}
        no_columns = self.model.shape[1]
        columns = list(range(no_columns-self.count, no_columns+1))
        selection = Selection([], [], [], columns, [])
        for key in selection.cell_generator(self.model.shape, self.grid.table):
            old_code = self.model.code_array(key)
            if old_code is not None:
                self.old_code[key] = old_code

        with self.model.inserting_columns(self.index, self.first, self.last):
            self.model.insertColumns(self.column, self.count)
        self.grid.table_choice.on_table_changed(self.grid.table)

    def undo(self):
        """Undo column insertion, updates screen"""

        # Clear must be first so that merged cells do not consume values
        self.model.code_array.dict_grid.cell_attributes.clear()

        with self.model.removing_rows(self.index, self.first, self.last):
            self.model.removeColumns(self.column, self.count)

        self.model.code_array.dict_grid.col_widths = self.old_col_widths

        for ca in self.old_cell_attributes:
            self.model.code_array.dict_grid.cell_attributes.append(ca)
        for key in self.old_code:
            self.model.code_array[key] = self.old_code[key]

        self.grid.table_choice.on_table_changed(self.grid.table)


class DeleteColumns(QUndoCommand):
    """Deletes grid columns"""

    def __init__(self, grid: QTableView, model: QAbstractTableModel,
                 index: QModelIndex, column: int, count: int,
                 description: str):
        """
        :param grid: The main grid object
        :param model: Model of the grid object
        :param index: Parent from which the new columns are deleted
        :param column: Column number of the first column to be deleted
        :param count: Number of columns to be deleted
        :param description: Command description

        """

        super().__init__(description)
        self.grid = grid
        self.model = model
        self.index = index
        self.column = column
        self.first = self.column = column
        self.last = column + count
        self.count = count

    def redo(self):
        """Redo column deletion, updates screen"""

        # Store content of deleted columns
        self.old_col_widths = copy(self.model.code_array.col_widths)
        self.old_cell_attributes = copy(self.model.code_array.cell_attributes)
        self.old_code = {}
        columns = list(range(self.first, self.last+1))
        selection = Selection([], [], [], columns, [])
        for key in selection.cell_generator(self.model.shape, self.grid.table):
            self.old_code[key] = self.model.code_array(key)

        with self.model.removing_columns(self.index, self.first, self.last):
            self.model.removeColumns(self.column, self.count)
        self.grid.table_choice.on_table_changed(self.grid.table)

    def undo(self):
        """Undo column deletion, updates screen"""

        # Clear must be first so that merged cells do not consume values
        self.model.code_array.dict_grid.cell_attributes.clear()

        with self.model.inserting_columns(self.index, self.first, self.last):
            self.model.insertColumns(self.column, self.count)

        self.model.code_array.dict_grid.col_widths = self.old_col_widths

        for ca in self.old_cell_attributes:
            self.model.code_array.dict_grid.cell_attributes.append(ca)
        for key in self.old_code:
            self.model.code_array[key] = self.old_code[key]

        self.grid.table_choice.on_table_changed(self.grid.table)


class InsertTable(QUndoCommand):
    """Inserts table"""

    def __init__(self, grid: QTableView, model: QAbstractTableModel,
                 table: int, description: str):
        """
        :param grid: The main grid object
        :param model: Model of the grid object
        :param table: Table number for insertion
        :param description: Command description

        """

        super().__init__(description)
        self.grid = grid
        self.model = model
        self.table = table

    def redo(self):
        """Redo table insertion, updates row and column sizes and screen"""

        # Store content of overflowing table
        self.old_row_heights = copy(self.model.code_array.row_heights)
        self.old_col_widths = copy(self.model.code_array.col_widths)
        self.old_cell_attributes = copy(self.model.code_array.cell_attributes)
        self.old_code = {}
        for key in self.model.code_array:
            if key[2] == self.model.shape[2] - 1:
                self.old_code[key] = self.model.code_array(key)

        with self.grid.undo_resizing_row():
            with self.grid.undo_resizing_column():
                self.model.insertTable(self.table)
        self.grid.table_choice.on_table_changed(self.grid.table)

    def undo(self):
        """Undo table insertion, updates row and column sizes and screen"""

        with self.grid.undo_resizing_row():
            with self.grid.undo_resizing_column():
                self.model.removeTable(self.table)

                self.model.code_array.dict_grid.row_heights = \
                    self.old_row_heights
                self.model.code_array.dict_grid.col_widths = \
                    self.old_col_widths
                self.model.code_array.dict_grid.cell_attributes.clear()
                for ca in self.old_cell_attributes:
                    self.model.code_array.dict_grid.cell_attributes.append(ca)
                for key in self.old_code:
                    self.model.code_array[key] = self.old_code[key]

        self.grid.table_choice.on_table_changed(self.grid.table)


class DeleteTable(QUndoCommand):
    """Deletes table"""

    def __init__(self, grid: QTableView, model: QAbstractTableModel,
                 table: int, description: str):
        """
        :param grid: The main grid object
        :param model: Model of the grid object
        :param table: Table number for deletion
        :param description: Command description

        """

        super().__init__(description)
        self.grid = grid
        self.model = model
        self.table = table

    def redo(self):
        """Redo table deletion, updates row and column sizes and screen"""

        # Store content of deleted table
        self.old_row_heights = copy(self.model.code_array.row_heights)
        self.old_col_widths = copy(self.model.code_array.col_widths)
        self.old_cell_attributes = copy(self.model.code_array.cell_attributes)
        self.old_code = {}
        for key in self.model.code_array:
            if key[2] == self.table:
                self.old_code[key] = self.model.code_array(key)

        with self.grid.undo_resizing_row():
            with self.grid.undo_resizing_column():
                self.model.removeTable(self.table)
        self.grid.table_choice.on_table_changed(self.grid.table)

    def undo(self):
        """Undo table deletion, updates row and column sizes and screen"""

        with self.grid.undo_resizing_row():
            with self.grid.undo_resizing_column():
                self.model.insertTable(self.table)

                self.model.code_array.dict_grid.row_heights = \
                    self.old_row_heights
                self.model.code_array.dict_grid.col_widths = \
                    self.old_col_widths
                self.model.code_array.dict_grid.cell_attributes.clear()
                for ca in self.old_cell_attributes:
                    self.model.code_array.dict_grid.cell_attributes.append(ca)
                for key in self.old_code:
                    self.model.code_array[key] = self.old_code[key]

        self.grid.table_choice.on_table_changed(self.grid.table)


class SetCellFormat(QUndoCommand):
    """Sets cell format in grid

    Format is set for one given cell and a selection.

    """

    def __init__(self, attr: CellAttribute, model: QAbstractTableModel,
                 index: QModelIndex, selected_idx: Iterable[QModelIndex],
                 description: str):
        """
        :param attr: Cell format to be set
        :param model: Model of the grid object
        :param index: Index of the cell for which the format is set
        :param selected_idx: Indexes of cells for which the format is set
        :param description: Command description

        """

        super().__init__(description)

        self.attr = attr
        self.model = model
        self.index = index
        self.selected_idx = selected_idx

    def redo(self):
        """Redo cell formatting"""

        self.model.setData(self.selected_idx, self.attr, Qt.DecorationRole)
        self.model.dataChanged.emit(QModelIndex(), QModelIndex())

    def undo(self):
        """Undo cell formatting"""

        self.model.code_array.cell_attributes.pop()
        self.model.dataChanged.emit(QModelIndex(), QModelIndex())


class SetCellMerge(SetCellFormat):
    """Sets cell merges in grid"""

    def redo(self):
        """Redo cell merging"""

        self.model.setData(self.selected_idx, self.attr, Qt.DecorationRole)
        for grid in self.model.main_window.grids:
            grid.update_cell_spans()
        self.model.dataChanged.emit(QModelIndex(), QModelIndex())

    def undo(self):
        """Undo cell merging"""

        try:
            self.model.code_array.cell_attributes.pop()
        except IndexError as error:
            raise Warning(str(error))
            return
        for grid in self.model.main_window.grids:
            grid.update_cell_spans()
        self.model.dataChanged.emit(QModelIndex(), QModelIndex())


class SetCellTextAlignment(SetCellFormat):
    """Sets cell text alignment in grid"""

    def redo(self):
        """Redo cell text alignment"""

        self.model.setData(self.selected_idx, self.attr, Qt.TextAlignmentRole)
        self.model.dataChanged.emit(QModelIndex(), QModelIndex())


class FreezeCell(QUndoCommand):
    """Freezes cell in grid"""

    def __init__(self, model: QAbstractTableModel,
                 cells: List[Tuple[int, int, int]], description: str):
        """
        :param model: Model of the grid object
        :param cells: List of indices of cells to be frozen
        :param description: Command description

        """

        super().__init__(description)

        self.model = model
        self.cells = cells

    def redo(self):
        """Redo cell freezing"""

        for cell in self.cells:
            row, column, table = cell

            # Add frozen cache content
            res_obj = self.model.code_array[cell]
            self.model.code_array.frozen_cache[repr(cell)] = res_obj

            # Set the frozen state
            selection = Selection([], [], [], [], [(row, column)])
            attr_dict = AttrDict([("frozen", True)])
            attr = CellAttribute(selection, table, attr_dict)
            self.model.setData([], attr, Qt.DecorationRole)
            self.model.dataChanged.emit(QModelIndex(), QModelIndex())

    def undo(self):
        """Undo cell freezing"""

        for cell in reversed(self.cells):
            self.model.code_array.frozen_cache.pop(repr(cell))
            self.model.code_array.cell_attributes.pop()
            self.model.dataChanged.emit(QModelIndex(), QModelIndex())


class ThawCell(FreezeCell):
    """Thaw (unfreezes) cell in grid"""

    def redo(self):
        """Redo cell thawing"""

        self.res_objs = []

        for cell in self.cells:
            row, column, table = cell

            if repr(cell) in self.model.code_array.frozen_cache:
                # Remove and store frozen cache content
                self.res_objs.append(
                    self.model.code_array.frozen_cache.pop(repr(cell)))

                # Remove the frozen state
                selection = Selection([], [], [], [], [(row, column)])
                attr_dict = AttrDict([("frozen", False)])
                attr = CellAttribute(selection, table, attr_dict)
                self.model.setData([], attr, Qt.DecorationRole)
                self.model.dataChanged.emit(QModelIndex(), QModelIndex())

    def undo(self):
        """Undo cell thawing"""

        for cell, res_obj in zip(reversed(self.cells),
                                 reversed(self.res_objs)):
            self.model.code_array.frozen_cache[repr(cell)] = res_obj
            self.model.code_array.cell_attributes.pop()
            self.model.dataChanged.emit(QModelIndex(), QModelIndex())


class SetCellRenderer(QUndoCommand):
    """Sets cell renderer in grid

    Adjusts syntax highlighting in entry line.

    """

    def __init__(self, attr: CellAttribute, model: QAbstractTableModel,
                 entry_line: QPlainTextEdit,
                 highlighter_document: QTextDocument,
                 index: QModelIndex, selected_idx: Iterable[QModelIndex],
                 description: str):
        """
        :param attr: Cell format that cointains target renderer information
        :param model: Model of the grid object
        :param entry_line: Entry line in main window
        :param highlighter_document: Document for entry line
        :param index: Index of the cell for which the renderer is set
        :param selected_idx: Indexes of cells for which the renderer is set
        :param description: Command description

        """

        super().__init__(description)

        self.attr = attr
        self.description = description
        self.model = model
        self.entry_line = entry_line
        self.new_highlighter_document = highlighter_document
        self.old_highlighter_document = self.entry_line.highlighter.document()
        self.index = index
        self.selected_idx = selected_idx

    def redo(self):
        """Redo cell renderer setting, adjusts syntax highlighting"""

        self.model.setData(self.selected_idx, self.attr, Qt.DecorationRole)
        self.entry_line.highlighter.setDocument(self.new_highlighter_document)
        self.model.dataChanged.emit(self.index, self.index)

    def undo(self):
        """Undo cell renderer setting, adjusts syntax highlighting"""

        self.model.code_array.cell_attributes.pop()
        self.entry_line.highlighter.setDocument(self.old_highlighter_document)
        self.model.dataChanged.emit(self.index, self.index)


class MakeButtonCell(QUndoCommand):
    """Makes a button cell"""

    def __init__(self, grid: QTableView, text: str, index: QModelIndex,
                 description: str):
        """
        :param grid: Main grid object
        :param text: Button cell text
        :param index: Index of the cell which becomes a button cell
        :param description: Command description

        """

        super().__init__(description)
        self.grid = grid
        self.text = text
        self.index = index
        self.key = self.index.row(), self.index.column(), self.grid.table

    def redo(self):
        """Redo button cell making"""

        row, column, table = self.key
        selection = Selection([], [], [], [], [(row, column)])
        attr_dict = AttrDict([("button_cell", self.text)])
        ca = CellAttribute(selection, table, attr_dict)
        self.grid.model.setData([self.index], ca, Qt.DecorationRole)

        if table == self.grid.table:
            # Only update widget if we are in the right table
            for grid in self.grid.main_window.grids:
                grid.update_index_widgets()

        self.grid.model.dataChanged.emit(self.index, self.index)

    def undo(self):
        """Undo button cell making"""

        row, column, table = self.key
        selection = Selection([], [], [], [], [(row, column)])
        attr_dict = AttrDict([("button_cell", False)])
        ca = CellAttribute(selection, table, attr_dict)
        self.grid.model.setData([self.index], ca, Qt.DecorationRole)

        if table == self.grid.table:
            # Only update widget if we are in the right table
            for grid in self.grid.main_window.grids:
                grid.update_index_widgets()

        self.grid.model.dataChanged.emit(self.index, self.index)


class RemoveButtonCell(QUndoCommand):
    """Removes a button cell"""

    def __init__(self, grid: QTableView, index: QModelIndex, description: str):
        """
        :param grid: Main grid object
        :param index: Index of the cell where a button cell is removed
        :param description: Command description

        """

        super().__init__(description)
        self.grid = grid
        self.text = None
        self.index = index
        self.key = self.index.row(), self.index.column(), self.grid.table

    def redo(self):
        """Redo button cell removal"""

        attr = self.grid.model.code_array.cell_attributes[self.key]
        self.text = attr.button_cell
        row, column, table = self.key
        selection = Selection([], [], [], [], [(row, column)])
        attr_dict = AttrDict([("button_cell", False)])
        ca = CellAttribute(selection, table, attr_dict)
        self.grid.model.setData([self.index], ca, Qt.DecorationRole)

        if table == self.grid.table:
            # Only update widget if we are in the right table
            for grid in self.grid.main_window.grids:
                grid.update_index_widgets()

        self.grid.model.dataChanged.emit(self.index, self.index)

    def undo(self):
        """Undo button cell removal"""

        row, column, table = self.key
        selection = Selection([], [], [], [], [(row, column)])
        attr_dict = AttrDict([("button_cell", self.text)])
        ca = CellAttribute(selection, table, attr_dict)
        self.grid.model.setData([self.index], ca, Qt.DecorationRole)

        if table == self.grid.table:
            # Only update widget if we are in the right table
            for grid in self.grid.main_window.grids:
                grid.update_index_widgets()

        self.grid.model.dataChanged.emit(self.index, self.index)



fig = Figure()
ax = fig.add_axes([.2,.2, .7, .7])

# Data for plotting
t = numpy.arange(0.0, 2.0, 0.01)
s = 1 + numpy.sin(2 * numpy.pi * t)

ax.plot(t, s)

ax.set(xlabel='Time (s)', ylabel='Voltage (mV)',
       title='Line chart')
ax.grid()

fig



# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------


"""

**Unit tests for string_helpers.py**

"""

import pytest
from ..string_helpers import quote, wrap_text


param_test_quote = [
    (None, None),
    (1, 1),
    ("", ""),
    ("Test", "'Test'"),
    ("ü+", "'ü+'"),
    (u"ü+", "'ü+'"),
    (r"ü+", "'ü+'"),
    (b"Test", "b'Test'"),
    ("Test1\nTest2", "'Test1\\nTest2'"),
]


@pytest.mark.parametrize("code, res", param_test_quote)
def test_quote(code, res):
    """Unit test for quote"""

    assert quote(code) == res


param_test_wrap_text = [
    ("", 80, 2000, ""),
    ("."*81, 80, 2000, "."*80+"\n."),
    (r"."*81, 80, 2000, "."*80+"\n."),
    ("~"*81, 80, 2000, "~"*80+"\n~"),
    (u"\u2200"*81, 80, 2000, "\u2200"*80+"\n\u2200"),
    ("."*160, 80, 2000, "."*80+"\n"+"."*80),
    ("x"*160, 80, 2, "xx..."),
    ("."*10, 2, 2000, "\n".join([".."]*5)),
    (None, 2, 10, None)
]


@pytest.mark.parametrize("text, width, maxlen, res", param_test_wrap_text)
def test_wrap_text(text, width, maxlen, res):
    """Unit test for wrap_text"""

    assert wrap_text(text, width, maxlen) == res



from matplotlib import cm

fig = Figure()
ax = fig.add_axes([.15,.15, .75, .75])

origin = 'lower'

delta = 0.025

x = y = numpy.arange(-3.0, 3.01, delta)
X, Y = numpy.meshgrid(x, y)
Z1 = numpy.exp(-X**2 - Y**2)
Z2 = numpy.exp(-(X - 1)**2 - (Y - 1)**2)
Z = (Z1 - Z2) * 2

CS = ax.contourf(X, Y, Z, 10, cmap=cm.bone, origin=origin)

CS2 = ax.contour(CS, levels=CS.levels[::2], colors='r', origin=origin)

ax.set_title('Contour chart', pad=15)
ax.set_xlabel('X')
ax.set_ylabel('Y')

# Make a colorbar for the ContourSet returned by the contourf call.
cbar = fig.colorbar(CS)
cbar.ax.set_ylabel('Z')

# Add the contour line levels to the colorbar
cbar.add_lines(CS2)

fig



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------


from setuptools import setup, find_packages

from pyspread import VERSION

with open("README.md", "r", encoding='utf8') as readme_file:
    long_description = readme_file.read()

setup(
    name='pyspread',
    version=VERSION,
    author='Martin Manns',
    author_email='mmanns@gmx.net',
    description='Pyspread is a non-traditional spreadsheet application'
    ' that is based on and written in the programming language Python.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://pyspread.gitlab.io",
    project_urls={
        "Bug Tracker": "https://gitlab.com/pyspread/pyspread/issues",
        "Documentation": "https://pyspread.gitlab.io/docs.html",
        "Source Code": "https://gitlab.com/pyspread/pyspread",
    },
    packages=find_packages(),
    entry_points={
        'console_scripts': {
            'pyspread = pyspread.pyspread:main'
        }
    },
    package_data={'pyspread': [
            'share/*',
            'share/*/*',
            'share/*/*/*',
            'share/*/*/*/*',
            'share/*/*/*/*/*',
        ]
    },
    data_files=[
        ('pyspread/share/applications',
         ['pyspread/share/applications/io.gitlab.pyspread.pyspread.desktop']),
    ],
    license='GPL v3 :: GNU General Public License',
    keywords=['spreadsheet', 'pyspread'],
    python_requires='>=3.6',
    install_requires=['numpy (>=1.1)',
                      'PyQt5 (>=5.10)',
                      'setuptools (>=40.0)',
                      'markdown2 (>=2.3)'],
    extras_require={
        'matplotlib': ['matplotlib (>=1.1.1)'],
        'pyenchant': ['pyenchant (>=1.1)'],
        'pip': ['pip (>=18)'],
        'python-dateutil': ['python-dateutil (>=2.7.0)'],
        'py-moneyed': ['py-moneyed (>=2.0)'],
        'rpy2': ['rpy2 (>=3.4)'],
        'plotnine': ['plotnine (>=0.8)'],
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: X11 Applications :: Qt',
        'Environment :: Win32 (MS Windows)',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Operating System :: POSIX',
        'Operating System :: Microsoft',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Office/Business :: Financial :: Spreadsheet',
        'Topic :: Scientific/Engineering',
    ],
)



# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------


"""
test_grid_renderer
==================

Unit tests for grid_renderer.py

"""

from contextlib import contextmanager
from os.path import abspath, dirname, join
import sys

import pytest

from PyQt5.QtWidgets import QApplication

PYSPREADPATH = abspath(join(dirname(__file__) + "/.."))
LIBPATH = abspath(PYSPREADPATH + "/lib")


@contextmanager
def insert_path(path):
    sys.path.insert(0, path)
    yield
    sys.path.pop(0)


with insert_path(PYSPREADPATH):
    from ..pyspread import MainWindow
    from ..grid_renderer import GridCellNavigator

app = QApplication.instance()
if app is None:
    app = QApplication([])
main_window = MainWindow()


class TestGridCellNavigator:
    """Unit tests for GridCellNavigator in grid_renderer.py"""

    grid = main_window.grid

    param_test_above_keys = [
        ((0, 0, 0), [(-1, 0, 0)]),
        ((20, 0, 0), [(19, 0, 0)]),
        ((20, 0, 2), [(19, 0, 2)]),
        ((20, 2, 2), [(19, 2, 2)]),
    ]

    @pytest.mark.parametrize("key, res", param_test_above_keys)
    def test_above_keys(self, key, res):
        """Unit test for above_keys"""

        cell = GridCellNavigator(self.grid, key)
        assert set(cell.above_keys()) == set(res)

    param_test_below_keys = [
        ((0, 0, 0), [(1, 0, 0)]),
        ((1000, 0, 0), [(1001, 0, 0)]),
        ((20, 0, 2), [(21, 0, 2)]),
        ((20, 2, 2), [(21, 2, 2)]),
    ]

    @pytest.mark.parametrize("key, res", param_test_below_keys)
    def test_below_keys(self, key, res):
        """Unit test for below_keys"""

        cell = GridCellNavigator(self.grid, key)
        assert set(cell.below_keys()) == set(res)

    param_test_left_keys = [
        ((0, 0, 0), [(0, -1, 0)]),
        ((20, 0, 0), [(20, -1, 0)]),
        ((20, 0, 2), [(20, -1, 2)]),
        ((20, 2, 2), [(20, 1, 2)]),
    ]

    @pytest.mark.parametrize("key, res", param_test_left_keys)
    def test_left_keys(self, key, res):
        """Unit test for left_keys"""

        cell = GridCellNavigator(self.grid, key)
        assert set(cell.left_keys()) == set(res)

    param_test_right_keys = [
        ((0, 0, 0), [(0, 1, 0)]),
        ((20, 0, 0), [(20, 1, 0)]),
        ((20, 0, 2), [(20, 1, 2)]),
        ((20, 2, 2), [(20, 3, 2)]),
    ]

    @pytest.mark.parametrize("key, res", param_test_right_keys)
    def test_right_keys(self, key, res):
        """Unit test for right_keys"""

        cell = GridCellNavigator(self.grid, key)
        assert set(cell.right_keys()) == set(res)

    param_test_above_left_key = [
        ((0, 0, 0), (-1, -1, 0)),
        ((20, 0, 0), (19, -1, 0)),
        ((20, 0, 2), (19, -1, 2)),
        ((20, 2, 2), (19, 1, 2)),
    ]

    @pytest.mark.parametrize("key, res", param_test_above_left_key)
    def test_above_left_key(self, key, res):
        """Unit test for above_left_key"""

        cell = GridCellNavigator(self.grid, key)
        assert cell.above_left_key() == res

    param_test_above_right_key = [
        ((0, 0, 0), (-1, 1, 0)),
        ((20, 0, 0), (19, 1, 0)),
        ((20, 0, 2), (19, 1, 2)),
        ((20, 2, 2), (19, 3, 2)),
    ]

    @pytest.mark.parametrize("key, res", param_test_above_right_key)
    def test_above_right_key(self, key, res):
        """Unit test for above_right_key"""

        cell = GridCellNavigator(self.grid, key)
        assert cell.above_right_key() == res

    param_test_below_left_key = [
        ((0, 0, 0), (1, -1, 0)),
        ((20, 0, 0), (21, -1, 0)),
        ((20, 0, 2), (21, -1, 2)),
        ((20, 2, 2), (21, 1, 2)),
    ]

    @pytest.mark.parametrize("key, res", param_test_below_left_key)
    def test_below_left_key(self, key, res):
        """Unit test for below_left_key"""

        cell = GridCellNavigator(self.grid, key)
        assert cell.below_left_key() == res

    param_test_below_right_key = [
        ((0, 0, 0), (1, 1, 0)),
        ((20, 0, 0), (21, 1, 0)),
        ((20, 0, 2), (21, 1, 2)),
        ((20, 2, 2), (21, 3, 2)),
    ]

    @pytest.mark.parametrize("key, res", param_test_below_right_key)
    def test_below_right_key(self, key, res):
        """Unit test for below_right_key"""

        cell = GridCellNavigator(self.grid, key)
        assert cell.below_right_key() == res



# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

"""

**Provides**

 * :class:`Entryline`

"""

from contextlib import contextmanager

from PyQt5.QtCore import Qt, QEvent
from PyQt5.QtGui import QTextOption, QKeyEvent
from PyQt5.QtWidgets import QWidget, QMainWindow

try:
    import pyspread.commands as commands
    from pyspread.lib.spelltextedit import SpellTextEdit
    from pyspread.lib.string_helpers import quote
except ImportError:
    import commands
    from lib.spelltextedit import SpellTextEdit
    from lib.string_helpers import quote


class Entryline(SpellTextEdit):
    """The entry line for pyspread"""

    def __init__(self, main_window: QMainWindow):
        """

        :param main_window: Application main window

        """

        font_family = main_window.settings.entry_line_font_family
        super().__init__(line_numbers=False, font_family=font_family)

        self.main_window = main_window

        min_height = self.cursorRect().y() + self.cursorRect().height() + 20
        self.setMinimumHeight(min_height)

        self.setWordWrapMode(QTextOption.WrapAnywhere)

        self.installEventFilter(self)

        self.last_key = None

        # self.highlighter.setDocument(self.document())

    # Overrides

    def eventFilter(self, source: QWidget, event: QEvent):
        """Quotes editor content for <Ctrl>+<Enter> and <Ctrl>+<Return>

        Overrides SpellTextEdit default shortcut. Counts as undoable action.

        :param source: Source widget of event
        :param event: Event to be filtered

        """

        if event.type() == QEvent.ShortcutOverride \
           and event.modifiers() == Qt.ControlModifier \
           and source == self \
           and event.key() in (Qt.Key_Return, Qt.Key_Enter):

            focused_grid = self.main_window.focused_grid
            code = quote(source.toPlainText())

            index = focused_grid.currentIndex()
            description = "Quote code for cell {}".format(index)
            cmd = commands.SetCellCode(code, focused_grid.model, index,
                                       description)
            self.main_window.undo_stack.push(cmd)
            focused_grid.setFocus()
            focused_grid.selectionModel().clearSelection()
            focused_grid.current = (focused_grid.row + 1, focused_grid.column,
                                    focused_grid.table)
            return False

        return QWidget.eventFilter(self, source, event)

    @contextmanager
    def disable_updates(self):
        """Disables updates and highlighter"""

        doc = self.highlighter.document()
        self.highlighter.setDocument(None)
        self.main_window.entry_line.setUpdatesEnabled(False)
        yield
        self.main_window.entry_line.setUpdatesEnabled(True)
        self.highlighter.setDocument(doc)

    def keyPressEvent(self, event: QKeyEvent):
        """Key press event filter

        :param event: Key event

        """

        grid = self.main_window.focused_grid

        self.last_key = event.key()

        if self.last_key in (Qt.Key_Enter, Qt.Key_Return):
            if event.modifiers() == Qt.ShiftModifier:
                self.insertPlainText('\n')
            else:
                if grid.selection_mode:
                    grid.set_selection_mode(False)
                else:
                    self.store_data()
                    grid.row += 1
        elif self.last_key == Qt.Key_Tab:
            self.store_data()
            grid.column += 1
        elif self.last_key == Qt.Key_Escape:
            grid.on_current_changed()
            grid.setFocus()
        elif self.last_key == Qt.Key_Insert:
            grid.set_selection_mode(not grid.selection_mode)
        else:
            super().keyPressEvent(event)

    def store_data(self):
        """Stores current entry line data in grid model"""

        grid = self.main_window.focused_grid

        index = grid.currentIndex()
        model = grid.model

        description = "Set code for cell {}".format(index)
        command = commands.SetCellCode(self.toPlainText(), model, index,
                                       description)
        self.main_window.undo_stack.push(command)

    def on_toggle_spell_check(self, signal: bool):
        """Spell check toggle event handler

        :param signal: Spell check is enabled if True

        """

        self.highlighter.enable_enchant = bool(signal)

    def setPlainText(self, text: str):
        """Overides setPlainText

        Additionally shows busy cursor and disables highlighter on long texts,
        and omits identical replace.

        :param text: Text to be set

        """

        is_long = (text is not None
                   and len(text) > self.main_window.settings.highlighter_limit)

        if text == self.toPlainText():
            return

        if is_long:
            with self.main_window.workflows.busy_cursor():
                self.highlighter.setDocument(None)
                super().setPlainText(text)
        else:
            super().setPlainText(text)



# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

"""

File hashing services

**Provides**

 * :func:`genkey` - Generates hash key
 * :func:`sign` - Returns a signature for a given file
 * :func:`verify` - Verifies file against signature

"""

import ast

from hashlib import blake2b
from hmac import compare_digest
import secrets


def genkey(nbytes: int = 64) -> bytes:
    """Returns a random byte sting that may be used as signature key

    :param nbytes: Length of key
    :return: Random byte string of length nbytes

    """

    return secrets.token_bytes(nbytes)


def sign(data: bytes, key: bytes) -> bytes:
    """Returns signature for file using blake2b

    Note: 64 bytes is the maximum that is supported in Python's BLAKE2b

    :param data: Data to be signed
    :param key: Signature key, len(key) <= 64
    :return: File signature hexdigest, encoded in utf-8

    """

    if not key:
        raise ValueError("No signature key defined")

    if not isinstance(key, bytes):
        key = ast.literal_eval(key)

    if len(key) > blake2b.MAX_KEY_SIZE:
        key = key[:blake2b.MAX_KEY_SIZE]
        raise UserWarning("Key is too long and has been truncated")

    signature = blake2b(digest_size=64, key=key)
    signature.update(data)

    return signature.hexdigest().encode('utf-8')


def verify(data: bytes, signature: bytes, key: bytes) -> bool:
    """Verifies a signature

    :param data: Data to be verified
    :param signature: Signature for verification
    :param key: Signature key, len(key) <= 64
    :return: True if verification was successful else False

    """

    data_signature = sign(data, key)
    return compare_digest(data_signature, signature)



from rpy2 import robjects
from rpy2.interactive import process_revents
from rpy2.robjects import Formula
from rpy2.robjects.packages import importr, data
from rpy2.robjects.lib import grdevices

try:
    process_revents.start()
except:
    pass

base = importr('base')
lattice = importr('lattice')
datasets = importr('datasets')

rprint = robjects.globalenv.find("print")

mtcars = data(datasets).fetch('mtcars')['mtcars']
formula = Formula('mpg ~ wt')
formula.getenvironment()['mpg'] = mtcars.rx2('mpg')
formula.getenvironment()['wt'] = mtcars.rx2('wt')
plot = lattice.xyplot(formula, type=base.c("p"), groups=mtcars.rx2('cyl'))

with grdevices.render_to_bytesio(grdevices.svg) as svg:
    rprint(plot)
process_revents.stop()

svg.getvalue() 



# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

"""

**Provides**

* :class:`IconPath`
* :class:`IconConverter`
* :class:`Icon`

"""


from PyQt5.QtGui import QIcon

try:
    from pyspread.settings import (ICON_PATH, ACTION_PATH, STATUS_PATH,
                                   CHARTS_PATH)
except ImportError:
    from settings import ICON_PATH, ACTION_PATH, STATUS_PATH, CHARTS_PATH


class IconPath:
    """Holds icon paths as attributes"""

    pyspread = ICON_PATH / 'hicolor' / 'svg' / 'pyspread.svg'

    # Status icons
    safe_mode = STATUS_PATH / 'status-safe-mode.svg'
    selection_mode = STATUS_PATH / 'status-selection-mode.svg'

    # File menu icons
    new = ACTION_PATH / 'document-new.svg'
    open = ACTION_PATH / 'document-open.svg'
    save = ACTION_PATH / 'document-save.svg'
    save_as = ACTION_PATH / 'document-save-as.svg'
    imprt = ACTION_PATH / 'document-import.svg'
    export = ACTION_PATH / 'document-export.svg'
    approve = ACTION_PATH / 'document-approve.svg'
    clear_globals = ACTION_PATH / 'edit-clear.svg'
    page_setup = ACTION_PATH / 'document-page-setup.svg'
    print_preview = ACTION_PATH / 'document-print-preview.svg'
    print = ACTION_PATH / 'document-print.svg'
    preferences = ACTION_PATH / 'document-properties.svg'
    new_gpg_key = ACTION_PATH / 'document-new-gpg-key.svg'
    quit = ACTION_PATH / 'document-log-out.svg'

    # Edit menu icons
    undo = ACTION_PATH / 'edit-undo.svg'
    redo = ACTION_PATH / 'edit-redo.svg'
    cut = ACTION_PATH / 'edit-cut.svg'
    copy = ACTION_PATH / 'edit-copy.svg'
    copy_results = ACTION_PATH / 'edit-copy-results.svg'
    paste = ACTION_PATH / 'edit-paste.svg'
    paste_as = ACTION_PATH / 'edit-paste-as.svg'
    select_all = ACTION_PATH / 'edit-select-all.svg'
    find = ACTION_PATH / 'edit-find.svg'
    find_next = ACTION_PATH / 'edit-find-next.svg'
    replace = ACTION_PATH / 'edit-find-replace.svg'
    sort_ascending = ACTION_PATH / 'edit-sort-ascending.svg'
    sort_descending = ACTION_PATH / 'edit-sort-descending.svg'
    quote = ACTION_PATH / 'edit-quote.svg'
    sort_ascending = ACTION_PATH / 'edit-sort-ascending.svg'
    sort_descending = ACTION_PATH / 'edit-sort-descending.svg'
    insert_row = ACTION_PATH / 'edit-insert-row.svg'
    insert_column = ACTION_PATH / 'edit-insert-column.svg'
    insert_table = ACTION_PATH / 'edit-insert-table.svg'
    delete_row = ACTION_PATH / 'edit-delete-row.svg'
    delete_column = ACTION_PATH / 'edit-delete-column.svg'
    delete_table = ACTION_PATH / 'edit-delete-table.svg'
    resize_grid = ACTION_PATH / 'edit-resize-grid.svg'

    # View menu icons
    fullscreen = ACTION_PATH / 'view-fullscreen.svg'
    goto_cell = ACTION_PATH / 'view-goto-cell.svg'
    check_spelling = ACTION_PATH / 'view-check-spelling.svg'
    zoom_in = ACTION_PATH / 'view-zoom-in.svg'
    zoom_out = ACTION_PATH / 'view-zoom-out.svg'
    zoom_1 = ACTION_PATH / 'view-zoom-original.svg'
    refresh = ACTION_PATH / 'view-refresh.svg'
    toggle_periodic_updates = ACTION_PATH / 'view-timer.svg'
    show_frozen = ACTION_PATH / 'view-show-frozen.svg'

    # Format menu icons
    copy_format = ACTION_PATH / 'format-copy.svg'
    paste_format = ACTION_PATH / 'format-paste.svg'
    font_dialog = ACTION_PATH / 'format-font.svg'
    bold = ACTION_PATH / 'format-text-bold.svg'
    italics = ACTION_PATH / 'format-text-italic.svg'
    underline = ACTION_PATH / 'format-text-underline.svg'
    strikethrough = ACTION_PATH / 'format-text-strikethrough.svg'
    markup = ACTION_PATH / 'format-cell-markup.svg'
    image = ACTION_PATH / 'format-cell-image.svg'
    text = ACTION_PATH / 'format-cell-text.svg'
    matplotlib = ACTION_PATH / 'format-cell-chart.svg'
    line_color = ACTION_PATH / 'format-line-color.svg'
    text_color = ACTION_PATH / 'format-text-color.svg'
    background_color = ACTION_PATH / 'format-background-color.svg'
    rotate_0 = ACTION_PATH / 'format-cell-rotate-0.svg'
    rotate_90 = ACTION_PATH / 'format-cell-rotate-90.svg'
    rotate_180 = ACTION_PATH / 'format-cell-rotate-180.svg'
    rotate_270 = ACTION_PATH / 'format-cell-rotate-270.svg'
    justify_left = ACTION_PATH / 'format-justify-left.svg'
    justify_fill = ACTION_PATH / 'format-justify-fill.svg'
    justify_center = ACTION_PATH / 'format-justify-center.svg'
    justify_right = ACTION_PATH / 'format-justify-right.svg'
    align_top = ACTION_PATH / 'format-text-align-top.svg'
    align_center = ACTION_PATH / 'format-text-align-center.svg'
    align_bottom = ACTION_PATH / 'format-text-align-bottom.svg'
    border_menu = ACTION_PATH / 'format-borders-all.svg'
    format_borders_all = ACTION_PATH / 'format-borders-all.svg'
    format_borders_top = ACTION_PATH / 'format-borders-top.svg'
    format_borders_bottom = ACTION_PATH / 'format-borders-bottom.svg'
    format_borders_left = ACTION_PATH / 'format-borders-left.svg'
    format_borders_right = ACTION_PATH / 'format-borders-right.svg'
    format_borders_outer = ACTION_PATH / 'format-borders-outer.svg'
    format_borders_inner = ACTION_PATH / 'format-borders-inner.svg'
    format_borders_top_bottom = ACTION_PATH / 'format-borders-top-bottom.svg'
    format_borders = ACTION_PATH / 'format-borders-4.svg'
    format_borders_0 = ACTION_PATH / 'format-borders-0.svg'
    format_borders_1 = ACTION_PATH / 'format-borders-1.svg'
    format_borders_2 = ACTION_PATH / 'format-borders-2.svg'
    format_borders_4 = ACTION_PATH / 'format-borders-4.svg'
    format_borders_8 = ACTION_PATH / 'format-borders-8.svg'
    format_borders_16 = ACTION_PATH / 'format-borders-16.svg'
    format_borders_32 = ACTION_PATH / 'format-borders-32.svg'
    format_borders_64 = ACTION_PATH / 'format-borders-64.svg'

    freeze = ACTION_PATH / 'format-freeze.svg'
    lock = ACTION_PATH / 'format-lock.svg'
    button = ACTION_PATH / 'format-button.svg'
    merge_cells = ACTION_PATH / 'format-merge-cells.svg'

    # Macro menu icons
    insert_image = ACTION_PATH / 'macro-insert-image.svg'
    link_image = ACTION_PATH / 'macro-link-image.svg'
    insert_chart = ACTION_PATH / 'macro-insert-chart.svg'
    insert_sum = ACTION_PATH / 'macro-insert-sum.svg'

    # Help menu icons
    help = ACTION_PATH / 'help-browser.svg'
    tutorial = ACTION_PATH / 'help-tutorial.svg'
    faq = ACTION_PATH / 'help-faq.svg'
    dependencies = ACTION_PATH / 'help-dependencies.svg'

    # Toolbar icons
    menu_manager = ACTION_PATH / 'menu-manager.svg'

    # Chart dialog template icons
    chart_pie_1_1 = CHARTS_PATH / 'chart_pie_1_1.svg'
    chart_ring_1_1 = CHARTS_PATH / 'chart_ring_1_1.svg'
    chart_line_1_1 = CHARTS_PATH / 'chart_line_1_1.svg'
    chart_polar_1_1 = CHARTS_PATH / 'chart_polar_1_1.svg'
    chart_area_1_1 = CHARTS_PATH / 'chart_area_1_1.svg'
    chart_column_1_1 = CHARTS_PATH / 'chart_column_1_1.svg'
    chart_column_1_2 = CHARTS_PATH / 'chart_column_1_2.svg'
    chart_bar_1_3 = CHARTS_PATH / 'chart_bar_1_3.svg'
    chart_scatter_1_1 = CHARTS_PATH / 'chart_scatter_1_1.svg'
    chart_bubble_1_1 = CHARTS_PATH / 'chart_bubble_1_1.svg'
    chart_boxplot_2_2 = CHARTS_PATH / 'chart_boxplot_2_2.svg'
    chart_histogram_1_1 = CHARTS_PATH / 'chart_histogram_1_1.svg'
    chart_histogram_1_4 = CHARTS_PATH / 'chart_histogram_1_4.svg'
    chart_scatterhist_1_1 = CHARTS_PATH/'chart_scatterhist_1_1.svg'
    chart_matrix_1_1 = CHARTS_PATH / 'chart_matrix_1_1.svg'
    chart_contour_1_2 = CHARTS_PATH / 'chart_contour_1_2.svg'
    chart_surface_2_1 = CHARTS_PATH / 'chart_surface_2_1.svg'

    chart_plotnine_geom_bar_1_1 = \
        CHARTS_PATH / 'chart_plotnine_geom_bar_1_1.svg'

    chart_r_graphics_barplot_1_1 = \
        CHARTS_PATH / 'chart_r_graphics_barplot_1_1.svg'
    chart_r_ggplot2_geom_boxplot_1_2 = \
        CHARTS_PATH / 'chart_r_ggplot2_geom_boxplot_1_2.svg'
    chart_r_ggplot2_geom_point_1_1 = \
        CHARTS_PATH / 'chart_r_ggplot2_geom_point_1_1.svg'
    chart_r_lattice_xyplot_1_1 = \
        CHARTS_PATH / 'chart_r_lattice_xyplot_1_1.svg'
    chart_r_ggplot2_geom_density2d_1_2 = \
        CHARTS_PATH / 'chart_r_ggplot2_geom_density2d_1_2.svg'
    chart_r_lattice_wireframe_2_1 =\
        CHARTS_PATH / 'chart_r_lattice_wireframe_2_1.svg'


class IconConverter(type):
    """Meta class that provides QIcons for IconPaths icons"""

    def __getattr__(cls, name: str) -> QIcon:
        """Provides QIcons for icon names

        :param name: Icon name

        """

        return QIcon(str(getattr(IconPath, name)))


class Icon(metaclass=IconConverter):
    """Provides QIcons as attributes for all attributes of IconPaths"""
